package org.netkernel.rdbms.endpoint;

import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.RequestScopeClassLoader;
import org.netkernel.module.standard.endpoint.StandardTransreptorImpl;
import org.netkernel.rdbms.representation.ConnectionPoolAspect;
import org.netkernel.rdbms.representation.IAspectDBConnectionPool;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.impl.RequestResponseFieldsImpl;

import java.lang.ref.WeakReference;
import java.lang.reflect.Method;
import java.sql.*;
import java.util.*;
import java.util.logging.Logger;

public class ConfigToRDBMSConnectionTransreptor extends StandardTransreptorImpl implements IEndpointStateMeta
{
	private Map<ConnectionPoolAspect, Boolean> mCreatedConnections=new WeakHashMap<ConnectionPoolAspect, Boolean>();
	private List<Driver> mDriversToDeregister=Collections.synchronizedList(new ArrayList<Driver>());
	private Timer mTimer;

	public ConfigToRDBMSConnectionTransreptor()
	{	this.declareToRepresentation(ConnectionPoolAspect.class);
		this.declareThreadSafe();
	}

	public void postCommission(INKFRequestContext aContext) throws Exception
	{	super.postCommission(aContext);
		mTimer=new Timer();
	}

	@Override
	public void onTransrept(INKFRequestContext helper) throws Exception
	{	IHDSNode hdsConfig=helper.sourcePrimary(IHDSNode.class);
		String jdbcConnection=(String) hdsConfig.getFirstValue("/config/rdbms/jdbcConnection");

		String jdbcDriver=null;
		int poolSize=1;
		int defaultQueryTimeout=0;
		int queryTimeoutMethod=IAspectDBConnectionPool.TIMEOUT_METHOD_SET;
		String testQuery=null;
		boolean testOnAcquire=false;
		long testOnIdlePeriod=0;
		long blockOnOverload=-1;
		Properties props=new Properties();
		for (IHDSNode i : hdsConfig.getNodes("/config/rdbms/*"))
		{	String name=i.getName();
			String value=((String) i.getValue());
			if (value!=null)
			{	value=value.trim();
			}
			else
			{	value="";
			}
			if (name.equals("jdbcDriver"))
			{	jdbcDriver=value;
			}
			else if (name.equals("jdbcConnection"))
			{	jdbcConnection=value;
			}
			else if (name.equals("poolSize"))
			{	poolSize=Integer.parseInt(value);
			}
			else if (name.equals("defaultQueryTimeout"))
			{	defaultQueryTimeout=Integer.parseInt(value);
			}
			else if (name.equals("queryTimeoutMethod"))
			{	if (value.equals("TIMER"))
				{	queryTimeoutMethod=IAspectDBConnectionPool.TIMEOUT_METHOD_TIMER;
				}
			}
			else if (name.equals("connectionTestQuery"))
			{	testQuery=value;
			}
			else if (name.equals("connectionTestOnAcquire"))
			{	testOnAcquire=Boolean.parseBoolean(value);
			}
			else if (name.equals("connectionTestOnIdlePeriod"))
			{	testOnIdlePeriod=Long.parseLong(value);
			}
			else if (name.equals("blockOnOverload"))
			{	blockOnOverload=Long.parseLong(value);
			}
			else
			{	props.put(name, value);
			}
		}

		ConnectionPoolAspect cpa=new ConnectionPoolAspect(jdbcConnection, props, poolSize, defaultQueryTimeout, queryTimeoutMethod, testOnAcquire, testQuery, testOnIdlePeriod, blockOnOverload, hdsConfig);

		//use an existing one if it exists
		ConnectionPoolAspect cpa3=null;
		synchronized (mCreatedConnections)
		{	for (ConnectionPoolAspect cpa2 : mCreatedConnections.keySet())
			{	if (cpa2.equals(cpa))
				{	cpa3=cpa2;
					break;
				}
			}
			if (cpa3==null)
			{	cpa3=cpa;
				mCreatedConnections.put(cpa, null);
				if (testOnIdlePeriod>0)
				{	TimerTask tt=new OnIdleConnectionTestTask(cpa3);
					mTimer.schedule(tt, testOnIdlePeriod, testOnIdlePeriod);
				}
			}
		}

		INKFResponse resp=helper.createResponseFrom(cpa3);
		mCreatedConnections.put(cpa, null);

		// initialise JDBC driver
		boolean needsRegistering=true;
		for (Enumeration e=DriverManager.getDrivers(); needsRegistering && e.hasMoreElements();)
		{	Driver d=(Driver) e.nextElement();
			needsRegistering=!d.acceptsURL(jdbcConnection);
		}
		if (needsRegistering)
		{	ClassLoader cl=new RequestScopeClassLoader(helper.getKernelContext().getThisKernelRequest().getRequestScope());  //Use a superstack classloader
			Class c=cl.loadClass(jdbcDriver);
			Driver d=(Driver) c.newInstance();
			d=new DriverShim(d);
			mDriversToDeregister.add(d);
			DriverManager.registerDriver(d);
		}
	}

	public void onDecommissionEndpoint() throws Exception
	{	mTimer.cancel();

		Iterator i=mCreatedConnections.keySet().iterator();
		while (i.hasNext())
		{	ConnectionPoolAspect cpa=(ConnectionPoolAspect) i.next();
			cpa.cleanup();
		}
		//deregister drivers
		for (Driver d : mDriversToDeregister)
		{	try
			{	DriverManager.deregisterDriver(d);
			}
			catch (SQLException e)
			{	e.printStackTrace();
			}
		}
		mDriversToDeregister.clear();

		super.onDecommissionEndpoint();
	}

	public IRequestResponseFields getState()
	{	RequestResponseFieldsImpl state=new RequestResponseFieldsImpl(super.getState());
		HDSBuilder b=new HDSBuilder();
		for (Driver d : mDriversToDeregister)
		{	b.pushNode("driver");
			Class driverClass=((DriverShim) d).getDelegate().getClass();
			b.addNode("class", driverClass.getName());
			b.addNode("classloader", driverClass.getClassLoader().toString());
			b.addNode("version", Integer.toString(d.getMajorVersion())+"."+Integer.toString(d.getMinorVersion()));
			b.addNode("toString", d.toString());
			b.popNode();
		}
		state.put("drivers", b.getRoot());

		b=new HDSBuilder();
		synchronized (mCreatedConnections)
		{	for (ConnectionPoolAspect cpa : mCreatedConnections.keySet())
			{	cpa.writeState(b);
			}
		}
		state.put("pools", b.getRoot());
		return state;
	}
}

class OnIdleConnectionTestTask extends TimerTask
{	private final WeakReference<ConnectionPoolAspect> mConnectionPool;

	public OnIdleConnectionTestTask(ConnectionPoolAspect aConnectionPool)
	{	mConnectionPool=new WeakReference<ConnectionPoolAspect>(aConnectionPool);
	}

	public void run()
	{	ConnectionPoolAspect cpa=mConnectionPool.get();
		if (cpa!=null)
		{	cpa.testIdleConnections();
		}
		else
		{	this.cancel();
		}
	}
}


/**
 * Thanks to Nick Sayer nsayer @ kfu.com for the solution to
 * jdbcs inability to load driver from alternate classloaders
 */
class DriverShim implements Driver
{	private Driver driver;

	DriverShim(Driver d)
	{	this.driver=d;
	}

	public boolean acceptsURL(String u) throws SQLException
	{	return this.driver.acceptsURL(u);
	}

	public Connection connect(String u, Properties p) throws SQLException
	{	return this.driver.connect(u, p);
	}

	public int getMajorVersion()
	{	return this.driver.getMajorVersion();
	}

	public int getMinorVersion()
	{	return this.driver.getMinorVersion();
	}

	public DriverPropertyInfo[] getPropertyInfo(String u, Properties p) throws SQLException
	{	return this.driver.getPropertyInfo(u, p);
	}

	public boolean jdbcCompliant()
	{	return this.driver.jdbcCompliant();
	}

	public String toString()
	{	return this.driver.toString();
	}

	public Driver getDelegate()
	{	return driver;
	}

	public Logger getParentLogger() throws SQLFeatureNotSupportedException
	{	/*Use reflection to deal with Java 6 / 7 API change*/
		try
		{	Method m = driver.getClass().getDeclaredMethod("getParentLogger", new Class[] {});
			Object result = m.invoke(driver, new Object[] {});
			return (Logger)result;
		}
		catch(Exception e)
		{	throw new SQLFeatureNotSupportedException(e);				
		}
	}
}

