/******************************************************************************
 (c) Copyright 2002 - $Date: 2008/04/11 11:57:59 $ 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile: ConnectionPoolAspect.java,v $
 * Version:       $Name:  $ $Revision: 1.1 $
 * Last Modified: $Date: 2008/04/11 11:57:59 $
 *****************************************************************************/
package org.netkernel.rdbms.representation;

import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;

import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

/**
 * @author tab
 */
public class ConnectionPoolAspect implements IAspectTransactedDBConnectionPool
{	private final String mConnectionString;
	private final Properties mProperties;
	private final int mPoolSize;
	private final int mDefaultQueryTimeout;
	private final int mQueryTimeoutMethod;
	private final boolean mTestOnAcquire;
	private final String mTestQuery;
	private final long mTestOnIdlePeriod;
	private final long mBlockOnOverload;
	private final List<Connection> mAvailablePool=new ArrayList();
	private final List<Connection> mUsedPool=new ArrayList();
	private static final int RETRY_COUNT=2;
	private Semaphore mBlockingSemaphore;
	private IHDSNode mHDSConfigurationForEquality;

	/**
	 * Creates a new instance of ConnectionPoolAspect
	 */
	public ConnectionPoolAspect(String aConnection, Properties aProperties, int aPoolSize, int aDefaultQueryTimeout, int aQueryTimeoutMethod, boolean aTestOnAquire, String aTestQuery, long aTestOnIdlePeriod, long aBlockOnOverload, IHDSNode aHDSConfig)
					throws ClassNotFoundException, InstantiationException, IllegalAccessException
	{	mConnectionString=aConnection;
		mProperties=aProperties;
		mPoolSize=aPoolSize;
		mDefaultQueryTimeout=aDefaultQueryTimeout;
		mQueryTimeoutMethod=aQueryTimeoutMethod;
		mTestOnAcquire=aTestOnAquire;
		mTestQuery=aTestQuery;
		mTestOnIdlePeriod=aTestOnIdlePeriod;
		mBlockOnOverload=aBlockOnOverload;
		if (mBlockOnOverload>=0)
		{	mBlockingSemaphore=new Semaphore(mPoolSize,true);
		}
		mHDSConfigurationForEquality=aHDSConfig;
	}
	

	public boolean equals(Object aOther)
	{	boolean result=aOther instanceof ConnectionPoolAspect;
		if (result)
		{	ConnectionPoolAspect other=(ConnectionPoolAspect) aOther;
			result=mHDSConfigurationForEquality.equals(other.mHDSConfigurationForEquality);
		}
		return result;
	}

	public int hashCode()
	{	return mConnectionString.hashCode();
	}

	public synchronized void writeState(HDSBuilder b)
	{	b.pushNode("pool");
		b.addNode("connection", mConnectionString);
		b.addNode("size", mAvailablePool.size()+mUsedPool.size());
		b.addNode("maxSize", mPoolSize);
		b.addNode("maxSize", mPoolSize);
		b.addNode("defaultQueryTimeout",mDefaultQueryTimeout);
		b.addNode("queryTimeoutMethod",mQueryTimeoutMethod==IAspectDBConnectionPool.TIMEOUT_METHOD_SET?"TIMEOUT_METHOD_SET":"TIMEOUT_METHOD_TIMER");
		b.addNode("blockOnOverload",mBlockOnOverload);
		b.addNode("properties",mProperties);
		b.popNode();
	}

	@Override
	protected void finalize() throws Throwable
	{	cleanup();
	}

	public synchronized void cleanup()
	{	for (Iterator i=mAvailablePool.iterator(); i.hasNext();)
		{	Connection c=(Connection) i.next();
			try
			{	c.close();
			} catch (SQLException e)
			{	//SysLogger.log(SysLogger.WARNING, this, "Exception whilst closing database connection in pool ["+mConnectionString+"]");
				e.printStackTrace();
			}
		}
		mAvailablePool.clear();
		if (!mUsedPool.isEmpty())
		{	//SysLogger.log(SysLogger.WARNING, this, "Connections still marked as in use when finalizing pool ["+mConnectionString+"]");
			for (Iterator i=mUsedPool.iterator(); i.hasNext();)
			{	Connection c=(Connection) i.next();
				try
				{	c.close();
				} catch (SQLException e)
				{
				}
			}
			mUsedPool.clear();
		}
	}

	private Connection innerAcquireConnection() throws NKFException, SQLException
	{	if (mBlockingSemaphore==null)
		{	
			return innerNonBlockingAcquireConnection();
		}
		else
		{	return innerBlockingAcquireConnection();
		}
	}

	private Connection innerBlockingAcquireConnection() throws NKFException, SQLException
	{	Connection result;
		try
		{	boolean success;
			if (mBlockOnOverload==0)
			{	mBlockingSemaphore.acquire();
				success=true;
			}
			else
			{	success=mBlockingSemaphore.tryAcquire(mBlockOnOverload, TimeUnit.MILLISECONDS);
			}
			if (success)
			{	int availablePoolSize=mAvailablePool.size();
				if (availablePoolSize>0)
				{	result=(Connection) mAvailablePool.remove(availablePoolSize-1);
					mUsedPool.add(result);
				}
				else 
				{	result=innerCreateNewConnection();
					mUsedPool.add(result);
				}
			}
			else
			{	throw new NKFException("Too Many Concurrent Connections", "timeout waiting for connection for ["+mConnectionString, null);
			}
		}
		catch (InterruptedException e)
		{	throw new NKFException("Interrupted", null, e);
		}
		return result;
	}
	
	private synchronized Connection innerNonBlockingAcquireConnection() throws NKFException, SQLException
	{	Connection result;
		int availablePoolSize=mAvailablePool.size();
		if (availablePoolSize>0)
		{	result=(Connection) mAvailablePool.remove(availablePoolSize-1);
			mUsedPool.add(result);
		}
		else if (mUsedPool.size()<mPoolSize)
		{	result=innerCreateNewConnection();
			mUsedPool.add(result);
		}
		else
		{	throw new NKFException("Too Many Concurrent Connections", ">"+mPoolSize+" connections requested for ["+mConnectionString, null);
		}
		return result;
	}

	private Connection innerCreateNewConnection() throws SQLException
	{	return DriverManager.getConnection(mConnectionString, mProperties);
	}

	private synchronized void innerReleaseConnection(Connection aConnection)
	{	if (mUsedPool.remove(aConnection))
		{	mAvailablePool.add(aConnection);
			if (mBlockingSemaphore!=null)
			{	mBlockingSemaphore.release();
			}
		}
	}

	public Connection acquireConnection() throws NKFException
	{	Connection result=null;
		int count=RETRY_COUNT;
		while (true)
		{	try
			{	result=innerAcquireConnection();
				if (mTestOnAcquire) testConnection(result);
				result.setAutoCommit(true);
				break;
			}
			catch (SQLException e)
			{	mUsedPool.remove(result);
				if (mBlockingSemaphore!=null)
				{	mBlockingSemaphore.release();
				}
				if ((count--)<=0)
				{	NKFException ex=new NKFException("Error Obtaining Connection", "on connection ["+mConnectionString+"]", null, e);
					throw ex;
				}
			}
		}
		return result;
	}

	private void testConnection(Connection aConnection) throws SQLException
	{	if (mTestQuery!=null)
		{	SQLException badTestException=null;
			Statement statement=aConnection.createStatement();
			try
			{	ResultSet rs=statement.executeQuery(mTestQuery);
				rs.close();
			}
			catch (SQLException e)
			{	badTestException=e;
				//SysLogger.log(SysLogger.WARNING, this, "Bad test-connection query ["+mTestQuery+"]");
			}
			if(badTestException!=null)
			{	try
				{	//Testing of connection failed - attempt to close it before returning and having it removed from pool.
					aConnection.close();
				}
				catch(Exception e2)
				{}
				throw badTestException;
			}
			else
			{	statement.close();
			}
		}
		else
		{	try
			{
				ResultSet rs=aConnection.getMetaData().getTables(null, null, null, null);
				rs.close();
			}
			catch(SQLException e)
			{	try
				{	//Testing of connection failed - attempt to close it before returning and having it removed from pool.
					aConnection.close();
				}
				catch(Exception e2)
				{}
				throw e;
			}
		}
	}

	public synchronized void testIdleConnections()
	{	//System.out.println("testIdleConnections() "+mConnectionString);
		for (Iterator<Connection> i=mAvailablePool.iterator(); i.hasNext();)
		{	Connection c=i.next();
			try
			{	testConnection(c);
			}
			catch (SQLException e)
			{	i.remove();
			}
		}
	}

	public void releaseConnection(Connection aConnection)
	{	innerReleaseConnection(aConnection);
	}

	public IAspectDBConnectionPool getTransactedConnection() throws NKFException
	{	TransactedConnection result;
		Connection c=null;
		int count=RETRY_COUNT;
		while (true)
		{	try
			{	c=innerAcquireConnection();
				if (mTestOnAcquire) testConnection(c);
				c.setAutoCommit(false);
				result=new TransactedConnection(c, this);
				break;
			}
			catch (SQLException e)
			{	mUsedPool.remove(c);
				if (mBlockingSemaphore!=null)
				{	mBlockingSemaphore.release();
				}
				if ((count--)<=0)
				{	NKFException ex=new NKFException("Error Obtaining Connection", "on connection ["+mConnectionString+"]", e);
					throw ex;
				}
			}
		}
		return result;
	}

	public void commitConnection(IAspectDBConnectionPool aConnection) throws NKFException
	{	if (aConnection instanceof TransactedConnection)
		{	TransactedConnection tc=(TransactedConnection) aConnection;
			Connection c=tc.acquireConnection();
			try
			{	c.commit();
			}
			catch (SQLException e)
			{	throw new NKFException("Error during Commit", "on connection ["+mConnectionString+"]", null);
			}
			finally
			{	releaseConnection(c);
			}
		}
		else
		{	//SysLogger.log(SysLogger.WARNING, this, "Bad commitConnection() in pool ["+mConnectionString+"]");
		}
	}

	public void rollbackConnection(IAspectDBConnectionPool aConnection) throws NKFException
	{	if (aConnection instanceof TransactedConnection)
		{	TransactedConnection tc=(TransactedConnection) aConnection;
			Connection c=tc.acquireConnection();
			try
			{	c.rollback();
			}
			catch (SQLException e)
			{	throw new NKFException("Error during Rollback", "on connection ["+mConnectionString+"]", null);
			}
			finally
			{	releaseConnection(c);
			}
		}
		else
		{	//SysLogger.log(SysLogger.WARNING, this, "Bad rollbackConnection() in pool ["+mConnectionString+"]");
		}
	}

	public int getDefaultQueryTimeout()
	{	return mDefaultQueryTimeout;
	}

	public int getQueryTimeoutMethod()
	{	return mQueryTimeoutMethod;
	}

	@Override
	public String toString()
	{	HDSBuilder b=new HDSBuilder();
		this.writeState(b);
		b.addNode("className", this.getClass().getName());
		return b.getRoot().toString();
	}
	
	

}
